import numpy as np

from keras_hub.src.models.esm.esm_backbone import ESMBackbone
from keras_hub.src.utils.preset_utils import get_file

backbone_cls = ESMBackbone


def convert_backbone_config(transformers_config):
    return {
        "vocabulary_size": transformers_config["vocab_size"],
        "num_layers": transformers_config["num_hidden_layers"],
        "num_heads": transformers_config["num_attention_heads"],
        "hidden_dim": transformers_config["hidden_size"],
        "intermediate_dim": transformers_config["intermediate_size"],
        "dropout": transformers_config["hidden_dropout_prob"],
        "position_embedding_type": transformers_config[
            "position_embedding_type"
        ],
        "pad_token_id": transformers_config["pad_token_id"],
        "max_sequence_length": transformers_config.get(
            "max_position_embeddings", None
        ),
        "layer_norm_eps": transformers_config.get("layer_norm_eps", 1e-12),
        "use_pre_layer_norm": transformers_config.get(
            "emb_layer_norm_before", False
        ),
        "activation": transformers_config.get("activation", "gelu"),
        "max_wavelength": transformers_config.get("max_wavelength", 10000),
    }


def transpose_and_reshape(x, shape):
    return np.reshape(np.transpose(x), shape)


def convert_weights(backbone, loader, transformers_config):
    # Embedding layer
    loader.port_weight(
        keras_variable=backbone.get_layer("token_embedding").embeddings,
        hf_weight_key="embeddings.word_embeddings.weight",
    )
    if transformers_config["position_embedding_type"] == "absolute":
        loader.port_weight(
            keras_variable=backbone.get_layer(
                "position_embedding"
            ).position_embeddings,
            hf_weight_key="embeddings.position_embeddings.weight",
        )
    if transformers_config.get("emb_layer_norm_before", False):
        loader.port_weight(
            keras_variable=backbone.get_layer("emb_layer_norm").gamma,
            hf_weight_key="embeddings.layer_norm.weight",
        )
        loader.port_weight(
            keras_variable=backbone.get_layer("emb_layer_norm").beta,
            hf_weight_key="embeddings.layer_norm.bias",
        )

    loader.port_weight(
        keras_variable=backbone.output_layer_norm.gamma,
        hf_weight_key="encoder.emb_layer_norm_after.weight",
    )
    loader.port_weight(
        keras_variable=backbone.output_layer_norm.beta,
        hf_weight_key="encoder.emb_layer_norm_after.bias",
    )

    # Attention blocks
    for i in range(backbone.num_layers):
        block = backbone.get_layer(f"transformer_layer_{i}")
        attn = block.attention_layer
        hf_prefix = "encoder.layer."
        # Attention layers
        loader.port_weight(
            keras_variable=attn.q_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.query.weight",
            hook_fn=transpose_and_reshape,
        )
        loader.port_weight(
            keras_variable=attn.q_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.query.bias",
            hook_fn=lambda hf_tensor, shape: np.reshape(hf_tensor, shape),
        )
        loader.port_weight(
            keras_variable=attn.k_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.key.weight",
            hook_fn=transpose_and_reshape,
        )
        loader.port_weight(
            keras_variable=attn.k_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.key.bias",
            hook_fn=lambda hf_tensor, shape: np.reshape(hf_tensor, shape),
        )
        loader.port_weight(
            keras_variable=attn.v_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.value.weight",
            hook_fn=transpose_and_reshape,
        )
        loader.port_weight(
            keras_variable=attn.v_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.value.bias",
            hook_fn=lambda hf_tensor, shape: np.reshape(hf_tensor, shape),
        )
        loader.port_weight(
            keras_variable=attn.o_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.attention.output.dense.weight",
            hook_fn=transpose_and_reshape,
        )
        loader.port_weight(
            keras_variable=attn.o_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.attention.output.dense.bias",
            hook_fn=lambda hf_tensor, shape: np.reshape(hf_tensor, shape),
        )
        # Attention layer norm.
        loader.port_weight(
            keras_variable=block.attention_norm.gamma,
            hf_weight_key=f"{hf_prefix}{i}.attention.LayerNorm.weight",
        )
        loader.port_weight(
            keras_variable=block.attention_norm.beta,
            hf_weight_key=f"{hf_prefix}{i}.attention.LayerNorm.bias",
        )
        # MLP layers
        loader.port_weight(
            keras_variable=block.feedforward_intermediate_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.intermediate.dense.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=block.feedforward_intermediate_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.intermediate.dense.bias",
        )
        loader.port_weight(
            keras_variable=block.feedforward_output_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.output.dense.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=block.feedforward_output_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.output.dense.bias",
        )
        # Output layer norm.
        loader.port_weight(
            keras_variable=block.feedforward_norm.gamma,
            hf_weight_key=f"{hf_prefix}{i}.LayerNorm.weight",
        )
        loader.port_weight(
            keras_variable=block.feedforward_norm.beta,
            hf_weight_key=f"{hf_prefix}{i}.LayerNorm.bias",
        )


def convert_tokenizer(cls, preset, **kwargs):
    return cls(
        get_file(preset, "vocab.txt"),
        lowercase=False,
        **kwargs,
    )



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.llama.llama_tokenizer import LlamaTokenizer


@keras_hub_export(
    [
        "keras_hub.tokenizers.MoonshineTokenizer",
        "keras_hub.models.MoonshineTokenizer",
    ]
)
class MoonshineTokenizer(LlamaTokenizer):
    """
    Moonshine tokenizer layer based on `keras_hub.models.LlamaTokenizer`.

    This tokenizer class is an alias of `LlamaTokenizer` but for the Moonshine
    model. It uses a SentencePiece vocabulary to handle tokenization.

    Args:
        proto: `str` or `bytes`. Either a string path to a SentencePiece proto
            file or a bytes object containing a serialized SentencePiece proto.
            See the [SentencePiece repository](https://github.com/google/sentencepiece)
            for details on the format.
        **kwargs: Additional keyword arguments passed to the parent
            `LlamaTokenizer`.

    Examples:
    ```python
    from keras_hub.tokenizers import MoonshineTokenizer

    # Initialize tokenizer.
    tokenizer = MoonshineTokenizer(
        "keras_hub/src/tests/test_data/llama_test_vocab.spm"
    )

    # Single input example.
    single_input = "the quick brown fox"
    single_tokens = tokenizer(single_input)
    print("Single input tokenization:")
    print(f"Input text: {single_input}")
    print(f"Tokenized: {single_tokens}")

    # Batched input example.
    batch_input = ["the quick brown fox", "the earth is round"]
    batch_tokens = tokenizer(batch_input)
    print("Batch input tokenization:")
    print(f"Input texts: {batch_input}")
    print(f"Tokenized: {batch_tokens}")

    # Detokenization example.
    encoded = tokenizer(single_input)
    decoded = tokenizer.detokenize(encoded)
    print("Detokenization:")
    print(f"Original text: {single_input}")
    print(f"Encoded: {encoded}")
    print(f"Decoded: {decoded}")
    ```
    """

    # NOTE: The 768 future-use tokens defined in Section 3.1 of the Moonshine
    # paper, "Moonshine: Speech Recognition for Live Transcription and Voice
    # Commands" (https://arxiv.org/pdf/2410.15608.pdf) serve no purpose in the
    # tokenizer at the moment, and are hence not included in the vocabulary.



import inspect

import keras
import numpy as np
from keras import ops

from keras_hub.src.layers.modeling.rotary_embedding import RotaryEmbedding
from keras_hub.src.models.gemma.rms_normalization import RMSNormalization
from keras_hub.src.utils.keras_utils import clone_initializer
from keras_hub.src.utils.keras_utils import fused_attention_op_available
from keras_hub.src.utils.keras_utils import gpu_supports_fused_attention_op
from keras_hub.src.utils.keras_utils import running_on_gpu
from keras_hub.src.utils.keras_utils import running_on_tpu


class CachedGemma3Attention(keras.layers.Layer):
    """A cached grouped query attention layer for Gemma3.

    This is the same as the attention layer used for Gemma and Gemma2. It
    exposes a few additional args:

    `use_query_key_norm`: bool. If True, apply RMS normalization on query
        and key. For Gemma3, this is True.
    `rope_wavelength`: float. Configurable value for RoPE wavelength. Gemma3
        uses 10K for local attention layers and 1M for global attention layers.
    `gate_dim_reduction`: int. In the gating layers, the output dimension is
        `intermediate_dim // gate_dim_reduction`. For Gemma and Gemma2, this
        value is 2. For Gemma3, it is 1.

    Moreover, the call() method takes in a `cache_update_mask` so as to make
    sure that the key-value cache is updated only for the non-prompt tokens
    during generation.
    """

    def __init__(
        self,
        head_dim,
        num_query_heads,
        num_key_value_heads,
        kernel_initializer="glorot_uniform",
        logit_soft_cap=None,
        use_sliding_window_attention=False,
        sliding_window_size=4096,
        query_head_dim_normalize=True,
        use_query_key_norm=False,
        layer_norm_epsilon=1e-6,
        rope_wavelength=10_000.0,
        rope_scaling_factor=1.0,
        use_bidirectional_attention=False,
        dropout=0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_query_heads = num_query_heads
        self.num_key_value_heads = num_key_value_heads
        self.head_dim = head_dim
        self.logit_soft_cap = logit_soft_cap
        self.use_sliding_window_attention = use_sliding_window_attention
        self.sliding_window_size = sliding_window_size
        self.query_head_dim_normalize = query_head_dim_normalize
        self.use_query_key_norm = use_query_key_norm
        self.layer_norm_epsilon = layer_norm_epsilon
        self.rope_wavelength = rope_wavelength
        self.rope_scaling_factor = rope_scaling_factor
        self.use_bidirectional_attention = use_bidirectional_attention
        self.dropout = dropout

        self._kernel_initializer = keras.initializers.get(
            clone_initializer(kernel_initializer)
        )
        self.num_key_value_groups = num_query_heads // num_key_value_heads
        self.query_head_dim_normalize = query_head_dim_normalize

    def build(self, inputs_shape):
        self.hidden_dim = inputs_shape[-1]

        self.query_dense = keras.layers.EinsumDense(
            "btd,ndh->btnh",
            output_shape=(None, self.num_query_heads, self.head_dim),
            kernel_initializer=self._kernel_initializer,
            dtype=self.dtype_policy,
            name="query",
        )
        self.query_dense.build(inputs_shape)

        self.key_dense = keras.layers.EinsumDense(
            "bsd,kdh->bskh",
            output_shape=(None, self.num_key_value_heads, self.head_dim),
            kernel_initializer=self._kernel_initializer,
            dtype=self.dtype_policy,
            name="key",
        )
        self.key_dense.build(inputs_shape)

        self.value_dense = keras.layers.EinsumDense(
            "bsd,kdh->bskh",
            output_shape=(None, self.num_key_value_heads, self.head_dim),
            kernel_initializer=self._kernel_initializer,
            dtype=self.dtype_policy,
            name="value",
        )
        self.value_dense.build(inputs_shape)

        if self.use_query_key_norm:
            self.query_norm = RMSNormalization(
                epsilon=self.layer_norm_epsilon,
                dtype=self.dtype_policy,
                name="query_norm",
            )
            self.query_norm.build(
                self.query_dense.compute_output_shape(inputs_shape)
            )

            self.key_norm = RMSNormalization(
                epsilon=self.layer_norm_epsilon,
                dtype=self.dtype_policy,
                name="key_norm",
            )
            self.key_norm.build(
                self.key_dense.compute_output_shape(inputs_shape)
            )

        self.dropout_layer = keras.layers.Dropout(
            rate=self.dropout,
            dtype=self.dtype_policy,
        )

        self.output_dense = keras.layers.EinsumDense(
            equation="btnh,nhd->btd",
            output_shape=(None, self.hidden_dim),
            kernel_initializer=self._kernel_initializer,
            dtype=self.dtype_policy,
            name="attention_output",
        )
        self.output_dense.build(
            (None, None, self.num_query_heads, self.head_dim)
        )
        self.softmax = keras.layers.Softmax(dtype="float32")

        self.rope_layer = RotaryEmbedding(
            max_wavelength=self.rope_wavelength,
            scaling_factor=self.rope_scaling_factor,
            dtype=self.dtype_policy,
        )

        self.built = True

    def _apply_rope(self, x, start_index):
        """Rope rotate q or k."""
        x = self.rope_layer(x, start_index=start_index)
        return x

    def _use_fused_attention_op(self):
        if not fused_attention_op_available():
            return False
        if self.dropout > 0.0:
            return False
        if running_on_gpu():
            # GPU never supports softcap in the fused op.
            if self.logit_soft_cap is not None:
                return False
            return gpu_supports_fused_attention_op()
        elif running_on_tpu():
            # TPU supports softcap with on keras >= 3.10.
            sig = inspect.signature(ops.dot_product_attention)
            return "attn_logits_soft_cap" in sig.parameters
        else:
            return False

    def _compute_attention(
        self,
        q,
        k,
        v,
        attention_mask,
        training=False,
        cache_update_index=0,
    ):
        if self.query_head_dim_normalize:
            query_normalization = 1 / np.sqrt(self.head_dim)
        else:
            query_normalization = 1 / np.sqrt(
                self.hidden_dim // self.num_query_heads
            )

        if self.use_sliding_window_attention and attention_mask is not None:
            attention_mask = self._mask_sliding_window(
                attention_mask,
                cache_update_index=cache_update_index,
            )

        if self._use_fused_attention_op():
            if attention_mask is not None:
                attention_mask = ops.expand_dims(attention_mask, axis=1)
                attention_mask = ops.cast(attention_mask, dtype="bool")
            # Only pass soft cap if needed as not all keras versions support.
            if self.logit_soft_cap:
                kwargs = {"attn_logits_soft_cap": self.logit_soft_cap}
            else:
                kwargs = {}
            return ops.dot_product_attention(
                query=q,
                key=k,
                value=v,
                mask=attention_mask,
                scale=query_normalization,
                **kwargs,
            )

        q *= ops.cast(query_normalization, dtype=q.dtype)
        q_shape = ops.shape(q)
        q = ops.reshape(
            q,
            (
                *q_shape[:-2],
                self.num_key_value_heads,
                self.num_query_heads // self.num_key_value_heads,
                q_shape[-1],
            ),
        )
        b, q_len, _, _, h = ops.shape(q)

        # Fallback to standard attention if flash attention is disabled
        attention_logits = ops.einsum("btkgh,bskh->bkgts", q, k)
        if self.logit_soft_cap is not None:
            attention_logits = ops.divide(attention_logits, self.logit_soft_cap)
            attention_logits = ops.multiply(
                ops.tanh(attention_logits), self.logit_soft_cap
            )

        if attention_mask is not None:
            attention_mask = attention_mask[:, None, None, :, :]
        orig_dtype = attention_logits.dtype
        attention_softmax = self.softmax(attention_logits, mask=attention_mask)
        attention_softmax = ops.cast(attention_softmax, orig_dtype)

        if self.dropout:
            attention_softmax = self.dropout_layer(
                attention_softmax, training=training
            )

        results = ops.einsum("bkgts,bskh->btkgh", attention_softmax, v)
        return ops.reshape(results, (b, q_len, self.num_query_heads, h))

    def _compute_bidirectional_sliding_mask(self, batch_size, sequence_length):
        """Computes a bidirectional sliding window attention mask.

        A token can attend to any other token if their absolute distance is
        within  half the sliding window size. This mask is used in embedding
        models like `EmbeddingGemma`.

        Args:
            batch_size: The batch size for the mask.
            sequence_length: The length of the sequence.

        Returns:
            A boolean attention mask with shape
            `(batch_size, sequence_length, sequence_length)`.
        """
        i = keras.ops.expand_dims(
            keras.ops.arange(sequence_length, dtype="int32"), axis=1
        )
        j = keras.ops.arange(sequence_length, dtype="int32")

        # If sliding window size is 4, the token in question attends to 1
        # token before and 2 tokens after.
        w_right = self.sliding_window_size // 2
        w_left = self.sliding_window_size - w_right - 1

        # Calculate the relative distance.
        distance = i - j

        mask = keras.ops.logical_and(distance <= w_left, distance >= -w_right)

        mask = keras.ops.expand_dims(mask, axis=0)
        return keras.ops.broadcast_to(
            mask, (batch_size, sequence_length, sequence_length)
        )

    def _mask_sliding_window(
        self,
        attention_mask,
        cache_update_index=0,
    ):
        batch_size, query_len, key_len = ops.shape(attention_mask)

        if self.use_bidirectional_attention:
            bidirectional_sliding_mask = (
                self._compute_bidirectional_sliding_mask(
                    batch_size=batch_size,
                    # `query_len = key_len` for embedding models
                    sequence_length=query_len,
                )
            )
            return ops.logical_and(attention_mask, bidirectional_sliding_mask)

        # Compute the sliding window for square attention.
        all_ones = ops.ones((key_len, key_len), "bool")
        if keras.config.backend() == "tensorflow":
            # TODO: trui/tril has issues with dynamic shape on the tensorflow
            # backend. We should fix, but use `band_part` for now.
            import tensorflow as tf

            band_size = ops.minimum(key_len, self.sliding_window_size - 1)
            band_size = ops.cast(band_size, "int32")
            sliding_mask = tf.linalg.band_part(all_ones, band_size, band_size)
        else:
            sliding_mask = ops.triu(
                all_ones, -1 * self.sliding_window_size + 1
            ) * ops.tril(all_ones, self.sliding_window_size - 1)
        # Slice the window for short queries during generation.
        start = (cache_update_index, 0)
        sliding_mask = ops.slice(sliding_mask, start, (query_len, key_len))
        sliding_mask = ops.expand_dims(sliding_mask, 0)
        return ops.logical_and(attention_mask, ops.cast(sliding_mask, "bool"))

    def call(
        self,
        x,
        attention_mask=None,
        cache=None,
        cache_update_index=0,
        cache_update_mask=None,
        training=False,
    ):
        query = self.query_dense(x)

        if self.use_query_key_norm:
            query = self.query_norm(query)

        query = self._apply_rope(query, cache_update_index)

        if cache is not None:
            key_cache = cache[:, 0, ...]
            value_cache = cache[:, 1, ...]
            key_update = self.key_dense(x)

            if self.use_query_key_norm:
                key_update = self.key_norm(key_update)

            key_update = self._apply_rope(key_update, cache_update_index)
            value_update = self.value_dense(x)

            # Update cache. Note that the cache is updated only if the
            # corresponding `cache_update_mask` value is True. This is to
            # ensure that we don't update the cache at indices corresponding to
            # the prompt. For Gemma3, in particular, this is useful because
            # image tokens have bidirectional attention. During generation,
            # if we have uneven inputs during generation, we might end up having
            # causal attention between image tokens, which is incorrect. To
            # avoid this, bidirectional attention is taken care of during
            # the prefill step, and during generation, the cache is not updated
            # for the prompt. The shape of `cache_update_mask` is
            # `(bsz, seq_len)`, where `seq_len` is 1 when we are generating
            # token-by-token.
            start = [0, cache_update_index, 0, 0]
            if cache_update_mask is not None:
                cache_update_mask = ops.expand_dims(
                    ops.expand_dims(cache_update_mask, axis=-1),
                    axis=-1,
                )
                key_original = ops.slice(
                    key_cache, start, ops.shape(key_update)
                )
                value_original = ops.slice(
                    value_cache, start, ops.shape(value_update)
                )

                key_update = ops.where(
                    cache_update_mask,
                    key_update,
                    key_original,
                )
                value_update = ops.where(
                    cache_update_mask,
                    value_update,
                    value_original,
                )

            key = ops.slice_update(key_cache, start, key_update)
            value = ops.slice_update(value_cache, start, value_update)
            cache = ops.stack((key, value), axis=1)
        else:
            key = self.key_dense(x)

            if self.use_query_key_norm:
                key = self.key_norm(key)

            key = self._apply_rope(key, cache_update_index)
            value = self.value_dense(x)

        attention_vec = self._compute_attention(
            query,
            key,
            value,
            attention_mask,
            training=training,
            cache_update_index=cache_update_index,
        )

        # Wipe attn vec if there are no attended tokens.
        no_attended_tokens = ops.all(
            ops.equal(attention_mask, 0), axis=-1, keepdims=True
        )[..., None]
        attention_vec = ops.where(
            no_attended_tokens, ops.zeros_like(attention_vec), attention_vec
        )

        attention_output = self.output_dense(attention_vec)

        if cache is not None:
            return attention_output, cache
        return attention_output

    def compute_output_shape(self, input_shape):
        return input_shape



import math

import keras
import numpy as np

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.d_fine.d_fine_decoder import DFineDecoder
from keras_hub.src.models.d_fine.d_fine_hybrid_encoder import DFineHybridEncoder
from keras_hub.src.models.d_fine.d_fine_layers import DFineAnchorGenerator
from keras_hub.src.models.d_fine.d_fine_layers import (
    DFineContrastiveDenoisingGroupGenerator,
)
from keras_hub.src.models.d_fine.d_fine_layers import (
    DFineInitialQueryAndReferenceGenerator,
)
from keras_hub.src.models.d_fine.d_fine_layers import DFineMLPPredictionHead
from keras_hub.src.models.d_fine.d_fine_layers import DFineSourceFlattener
from keras_hub.src.models.d_fine.d_fine_layers import (
    DFineSpatialShapesExtractor,
)
from keras_hub.src.models.d_fine.d_fine_utils import d_fine_kernel_initializer
from keras_hub.src.utils.keras_utils import standardize_data_format


class DFineDenoisingPreprocessorLayer(keras.layers.Layer):
    """Processes and prepares tensors for contrastive denoising.

    This layer is a helper used within the `DFineBackbone`'s functional model
    definition. Its primary role is to take the outputs from the
    `DFineContrastiveDenoisingGroupGenerator` and prepare them for the dynamic,
    per-batch forward pass, mostly since this functionality cannot be integrated
    directly into the `DFineBackbone` in the symbolic forward pass.

    The layer takes a tuple of `(pixel_values, input_query_class,
    denoising_bbox_unact, attention_mask)` and an optional
    `denoising_meta_values` dictionary as input to its `call` method.
    """

    def __init__(self, dtype=None, **kwargs):
        super().__init__(dtype=dtype, **kwargs)

    def call(self, inputs, denoising_meta_values=None):
        (
            pixel_values,
            input_query_class,
            denoising_bbox_unact,
            attention_mask,
        ) = inputs
        input_query_class_tensor = keras.ops.convert_to_tensor(
            input_query_class, dtype="int32"
        )
        denoising_bbox_unact_tensor = keras.ops.convert_to_tensor(
            denoising_bbox_unact, dtype=self.compute_dtype
        )
        attention_mask_tensor = keras.ops.convert_to_tensor(
            attention_mask, dtype=self.compute_dtype
        )
        outputs = {
            "input_query_class": input_query_class_tensor,
            "denoising_bbox_unact": denoising_bbox_unact_tensor,
            "attention_mask": attention_mask_tensor,
        }

        if denoising_meta_values is not None:
            batch_size = keras.ops.shape(pixel_values)[0]
            dn_positive_idx = denoising_meta_values["dn_positive_idx"]
            c_batch_size = keras.ops.shape(dn_positive_idx)[0]
            if c_batch_size == 0:
                outputs["dn_positive_idx"] = keras.ops.zeros(
                    (batch_size,) + keras.ops.shape(dn_positive_idx)[1:],
                    dtype=dn_positive_idx.dtype,
                )
            else:
                num_repeats = (batch_size + c_batch_size - 1) // c_batch_size
                dn_positive_idx_tiled = keras.ops.tile(
                    dn_positive_idx,
                    (num_repeats,)
                    + (1,) * (keras.ops.ndim(dn_positive_idx) - 1),
                )
                outputs["dn_positive_idx"] = dn_positive_idx_tiled[:batch_size]
            dn_num_group = denoising_meta_values["dn_num_group"]
            outputs["dn_num_group"] = keras.ops.tile(
                keras.ops.expand_dims(dn_num_group, 0), (batch_size,)
            )
            dn_num_split = denoising_meta_values["dn_num_split"]
            outputs["dn_num_split"] = keras.ops.tile(
                keras.ops.expand_dims(dn_num_split, 0), (batch_size, 1)
            )

        return outputs


@keras_hub_export("keras_hub.models.DFineBackbone")
class DFineBackbone(Backbone):
    """D-FINE Backbone for Object Detection.

    This class implements the core D-FINE architecture, which serves as the
    backbone for `DFineObjectDetector`. It integrates a `HGNetV2Backbone` for
    initial feature extraction, a `DFineHybridEncoder` for multi-scale feature
    fusion using FPN/PAN pathways, and a `DFineDecoder` for refining object
    queries.

    The backbone orchestrates the entire forward pass, from processing raw
    pixels to generating intermediate predictions. Key steps include:
    1.  Extracting multi-scale feature maps using the HGNetV2 backbone.
    2.  Fusing these features with the hybrid encoder.
    3.  Generating anchor proposals and selecting the top-k to initialize
        decoder queries and reference points.
    4.  Generating noisy queries for contrastive denoising (if the `labels`
        argument is provided).
    5.  Passing the queries and fused features through the transformer decoder
        to produce iterative predictions for bounding boxes and class logits.

    Args:
        backbone: A `keras.Model` instance that serves as the feature extractor.
            While any `keras.Model` can be used, we highly recommend using a
            `keras_hub.models.HGNetV2Backbone` instance, as this architecture is
            optimized for its outputs. If a custom backbone is provided, it
            must have a `stage_names` attribute, or the `out_features` argument
            for this model must be specified. This requirement helps prevent
            hard-to-debug downstream dimensionality errors.
        decoder_in_channels: list, Channel dimensions of the multi-scale
            features from the hybrid encoder. This should typically be a list
            of `encoder_hidden_dim` repeated for each feature level.
        encoder_hidden_dim: int, Hidden dimension size for the encoder layers.
        num_labels: int, Number of object classes for detection.
        num_denoising: int, Number of denoising queries for contrastive
            denoising training. Set to `0` to disable denoising.
        learn_initial_query: bool, Whether to learn initial query embeddings.
        num_queries: int, Number of object queries for detection.
        anchor_image_size: tuple, Size of the anchor image as `(height, width)`.
        feat_strides: list, List of feature stride values for different pyramid
            levels.
        num_feature_levels: int, Number of feature pyramid levels to use.
        hidden_dim: int, Hidden dimension size for the model.
        encoder_in_channels: list, Channel dimensions of the feature maps from
            the backbone (`HGNetV2Backbone`) that are fed into the hybrid
            encoder.
        encode_proj_layers: list, List specifying projection layer
            configurations.
        num_attention_heads: int, Number of attention heads in encoder layers.
        encoder_ffn_dim: int, Feed-forward network dimension in encoder.
        num_encoder_layers: int, Number of encoder layers.
        hidden_expansion: float, Hidden dimension expansion factor.
        depth_multiplier: float, Depth multiplier for the backbone.
        eval_idx: int, Index for evaluation. Defaults to `-1` for the last
            layer.
        num_decoder_layers: int, Number of decoder layers.
        decoder_attention_heads: int, Number of attention heads in decoder
            layers.
        decoder_ffn_dim: int, Feed-forward network dimension in decoder.
        decoder_method: str, Decoder method. Can be either `"default"` or
            `"discrete"`. Defaults to `"default"`.
        decoder_n_points: list, Number of sampling points for deformable
            attention.
        lqe_hidden_dim: int, Hidden dimension for learned query embedding.
        num_lqe_layers: int, Number of layers in learned query embedding.
        label_noise_ratio: float, Ratio of label noise for denoising
            training. Defaults to `0.5`.
        box_noise_scale: float, Scale factor for box noise in denoising
            training. Defaults to `1.0`.
        labels: list or None, Ground truth labels for denoising training. This
            is passed during model initialization to construct the training
            graph for contrastive denoising. Each element should be a
            dictionary with `"boxes"` (numpy array of shape `[N, 4]` with
            normalized coordinates) and `"labels"` (numpy array of shape `[N]`
            with class indices). Required when `num_denoising > 0`. Defaults to
            `None`.
        seed: int or None, Random seed for reproducibility. Defaults to `None`.
        image_shape: tuple, Shape of input images as `(height, width,
            channels)`. Height and width can be `None` for variable input sizes.
            Defaults to `(None, None, 3)`.
        out_features: list or None, List of feature names to output from
            backbone. If `None`, uses the last `len(decoder_in_channels)`
            features from the backbone's `stage_names`. Defaults to `None`.
        data_format: str, The data format of the image channels. Can be either
            `"channels_first"` or `"channels_last"`. If `None` is specified,
            it will use the `image_data_format` value found in your Keras
            config file at `~/.keras/keras.json`. Defaults to `None`.
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights. Defaults to `None`.
        **kwargs: Additional keyword arguments passed to the base class.

    Example:
    ```python
    import keras
    import numpy as np
    from keras_hub.models import DFineBackbone
    from keras_hub.models import HGNetV2Backbone

    # Example 1: Basic usage without denoising.
    # First, build the `HGNetV2Backbone` instance.
    hgnetv2 = HGNetV2Backbone(
        stem_channels=[3, 16, 16],
        stackwise_stage_filters=[
            [16, 16, 64, 1, 3, 3],
            [64, 32, 256, 1, 3, 3],
            [256, 64, 512, 2, 3, 5],
            [512, 128, 1024, 1, 3, 5],
        ],
        apply_downsample=[False, True, True, True],
        use_lightweight_conv_block=[False, False, True, True],
        depths=[1, 1, 2, 1],
        hidden_sizes=[64, 256, 512, 1024],
        embedding_size=16,
        use_learnable_affine_block=True,
        hidden_act="relu",
        image_shape=(None, None, 3),
        out_features=["stage3", "stage4"],
        data_format="channels_last",
    )

    # Then, pass the backbone instance to `DFineBackbone`.
    backbone = DFineBackbone(
        backbone=hgnetv2,
        decoder_in_channels=[128, 128],
        encoder_hidden_dim=128,
        num_denoising=0,  # Disable denoising
        num_labels=80,
        hidden_dim=128,
        learn_initial_query=False,
        num_queries=300,
        anchor_image_size=(256, 256),
        feat_strides=[16, 32],
        num_feature_levels=2,
        encoder_in_channels=[512, 1024],
        encode_proj_layers=[1],
        num_attention_heads=8,
        encoder_ffn_dim=512,
        num_encoder_layers=1,
        hidden_expansion=0.34,
        depth_multiplier=0.5,
        eval_idx=-1,
        num_decoder_layers=3,
        decoder_attention_heads=8,
        decoder_ffn_dim=512,
        decoder_n_points=[6, 6],
        lqe_hidden_dim=64,
        num_lqe_layers=2,
        out_features=["stage3", "stage4"],
        image_shape=(None, None, 3),
        data_format="channels_last",
        seed=0,
    )

    # Prepare input data.
    input_data = keras.random.uniform((2, 256, 256, 3))

    # Forward pass.
    outputs = backbone(input_data)

    # Example 2: With contrastive denoising training.
    labels = [
        {
            "boxes": np.array([[0.5, 0.5, 0.2, 0.2], [0.4, 0.4, 0.1, 0.1]]),
            "labels": np.array([1, 10]),
        },
        {
            "boxes": np.array([[0.6, 0.6, 0.3, 0.3]]),
            "labels": np.array([20]),
        },
    ]

    # Pass the `HGNetV2Backbone` instance to `DFineBackbone`.
    backbone_with_denoising = DFineBackbone(
        backbone=hgnetv2,
        decoder_in_channels=[128, 128],
        encoder_hidden_dim=128,
        num_denoising=100,  # Enable denoising
        num_labels=80,
        hidden_dim=128,
        learn_initial_query=False,
        num_queries=300,
        anchor_image_size=(256, 256),
        feat_strides=[16, 32],
        num_feature_levels=2,
        encoder_in_channels=[512, 1024],
        encode_proj_layers=[1],
        num_attention_heads=8,
        encoder_ffn_dim=512,
        num_encoder_layers=1,
        hidden_expansion=0.34,
        depth_multiplier=0.5,
        eval_idx=-1,
        num_decoder_layers=3,
        decoder_attention_heads=8,
        decoder_ffn_dim=512,
        decoder_n_points=[6, 6],
        lqe_hidden_dim=64,
        num_lqe_layers=2,
        out_features=["stage3", "stage4"],
        image_shape=(None, None, 3),
        seed=0,
        labels=labels,
    )

    # Forward pass with denoising.
    outputs_with_denoising = backbone_with_denoising(input_data)
    ```
    """

    def __init__(
        self,
        backbone,
        decoder_in_channels,
        encoder_hidden_dim,
        num_labels,
        num_denoising,
        learn_initial_query,
        num_queries,
        anchor_image_size,
        feat_strides,
        num_feature_levels,
        hidden_dim,
        encoder_in_channels,
        encode_proj_layers,
        num_attention_heads,
        encoder_ffn_dim,
        num_encoder_layers,
        hidden_expansion,
        depth_multiplier,
        eval_idx,
        num_decoder_layers,
        decoder_attention_heads,
        decoder_ffn_dim,
        decoder_n_points,
        lqe_hidden_dim,
        num_lqe_layers,
        decoder_method="default",
        label_noise_ratio=0.5,
        box_noise_scale=1.0,
        labels=None,
        seed=None,
        image_shape=(None, None, 3),
        out_features=None,
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        if decoder_method not in ["default", "discrete"]:
            decoder_method = "default"
        data_format = standardize_data_format(data_format)
        channel_axis = -1 if data_format == "channels_last" else 1
        self.backbone = backbone
        # Re-instantiate the backbone if its data_format mismatches the parents.
        if (
            hasattr(self.backbone, "data_format")
            and self.backbone.data_format != data_format
        ):
            backbone_config = self.backbone.get_config()
            backbone_config["data_format"] = data_format
            if (
                "image_shape" in backbone_config
                and backbone_config["image_shape"] is not None
                and len(backbone_config["image_shape"]) == 3
            ):
                backbone_config["image_shape"] = tuple(
                    reversed(backbone_config["image_shape"])
                )
            self.backbone = self.backbone.__class__.from_config(backbone_config)
        spatial_shapes = []
        for s in feat_strides:
            h = anchor_image_size[0] // s
            w = anchor_image_size[1] // s
            spatial_shapes.append((h, w))
        # NOTE: While `HGNetV2Backbone` is handled automatically, `out_features`
        # must be specified for custom backbones. This design choice prevents
        # hard-to-debug dimension mismatches by placing the onus on the user for
        # ensuring compatibility.
        if not hasattr(self.backbone, "stage_names") and out_features is None:
            raise ValueError(
                "`out_features` must be specified when using a custom "
                "backbone that does not have a `stage_names` attribute."
            )
        stage_names = getattr(self.backbone, "stage_names", out_features)
        out_features = (
            out_features
            if out_features is not None
            else stage_names[-len(decoder_in_channels) :]
        )
        initializer = d_fine_kernel_initializer(
            initializer_range=0.01,
        )

        # === Layers ===
        self.encoder = DFineHybridEncoder(
            encoder_in_channels=encoder_in_channels,
            feat_strides=feat_strides,
            encoder_hidden_dim=encoder_hidden_dim,
            encode_proj_layers=encode_proj_layers,
            positional_encoding_temperature=10000,
            eval_size=None,
            normalize_before=False,
            num_attention_heads=num_attention_heads,
            dropout=0.0,
            layer_norm_eps=1e-5,
            encoder_activation_function="gelu",
            activation_dropout=0.0,
            encoder_ffn_dim=encoder_ffn_dim,
            num_encoder_layers=num_encoder_layers,
            batch_norm_eps=1e-5,
            hidden_expansion=hidden_expansion,
            depth_multiplier=depth_multiplier,
            kernel_initializer=initializer,
            bias_initializer="zeros",
            channel_axis=channel_axis,
            data_format=data_format,
            dtype=dtype,
            name="hybrid_encoder",
        )
        self.decoder = DFineDecoder(
            layer_scale=1.0,
            eval_idx=eval_idx,
            num_decoder_layers=num_decoder_layers,
            dropout=0.0,
            hidden_dim=hidden_dim,
            reg_scale=4.0,
            max_num_bins=32,
            upsampling_factor=0.5,
            decoder_attention_heads=decoder_attention_heads,
            attention_dropout=0.0,
            decoder_activation_function="relu",
            activation_dropout=0.0,
            layer_norm_eps=1e-5,
            decoder_ffn_dim=decoder_ffn_dim,
            num_feature_levels=num_feature_levels,
            decoder_offset_scale=0.5,
            decoder_method=decoder_method,
            decoder_n_points=decoder_n_points,
            top_prob_values=4,
            lqe_hidden_dim=lqe_hidden_dim,
            num_lqe_layers=num_lqe_layers,
            num_labels=num_labels,
            spatial_shapes=spatial_shapes,
            dtype=dtype,
            initializer_bias_prior_prob=None,
            num_queries=num_queries,
            name="decoder",
        )
        self.anchor_generator = DFineAnchorGenerator(
            anchor_image_size=anchor_image_size,
            feat_strides=feat_strides,
            data_format=data_format,
            dtype=dtype,
            name="anchor_generator",
        )
        self.contrastive_denoising_group_generator = (
            DFineContrastiveDenoisingGroupGenerator(
                num_labels=num_labels,
                num_denoising=num_denoising,
                label_noise_ratio=label_noise_ratio,
                box_noise_scale=box_noise_scale,
                seed=seed,
                dtype=dtype,
                name="contrastive_denoising_group_generator",
            )
        )
        if num_denoising > 0:
            self.denoising_class_embed = keras.layers.Embedding(
                input_dim=num_labels + 1,
                output_dim=hidden_dim,
                embeddings_initializer="glorot_uniform",
                name="denoising_class_embed",
                dtype=dtype,
            )
            self.denoising_class_embed.build(None)
        else:
            self.denoising_class_embed = None

        self.source_flattener = DFineSourceFlattener(
            dtype=dtype,
            name="source_flattener",
            channel_axis=channel_axis,
            data_format=data_format,
        )
        self.initial_query_reference_generator = (
            DFineInitialQueryAndReferenceGenerator(
                num_queries=num_queries,
                learn_initial_query=learn_initial_query,
                hidden_dim=hidden_dim,
                dtype=dtype,
                name="initial_query_reference_generator",
            )
        )
        self.spatial_shapes_extractor = DFineSpatialShapesExtractor(
            dtype=dtype,
            data_format=data_format,
            name="spatial_shapes_extractor",
        )
        num_backbone_outs = len(decoder_in_channels)
        self.encoder_input_proj_layers = []
        for i in range(num_backbone_outs):
            self.encoder_input_proj_layers.append(
                [
                    keras.layers.Conv2D(
                        filters=encoder_hidden_dim,
                        kernel_size=1,
                        use_bias=False,
                        kernel_initializer=initializer,
                        bias_initializer="zeros",
                        data_format=data_format,
                        name=f"encoder_input_proj_conv_{i}",
                        dtype=dtype,
                    ),
                    keras.layers.BatchNormalization(
                        epsilon=1e-5,
                        axis=channel_axis,
                        name=f"encoder_input_proj_bn_{i}",
                        dtype=dtype,
                    ),
                ]
            )
        self.enc_output_layers = [
            keras.layers.Dense(
                hidden_dim,
                name="enc_output_dense",
                dtype=dtype,
            ),
            keras.layers.LayerNormalization(
                epsilon=1e-5,
                name="enc_output_ln",
                dtype=dtype,
            ),
        ]
        prior_prob = 1 / (num_labels + 1)
        enc_score_head_bias = float(-math.log((1 - prior_prob) / prior_prob))
        self.enc_score_head = keras.layers.Dense(
            num_labels,
            name="enc_score_head",
            dtype=dtype,
            kernel_initializer="glorot_uniform",
            bias_initializer=keras.initializers.Constant(enc_score_head_bias),
        )
        self.enc_bbox_head = DFineMLPPredictionHead(
            input_dim=hidden_dim,
            hidden_dim=hidden_dim,
            output_dim=4,
            num_layers=3,
            name="enc_bbox_head",
            dtype=dtype,
            kernel_initializer=initializer,
            last_layer_initializer="zeros",
        )
        self.decoder_input_proj_layers = []
        for i in range(num_backbone_outs):
            if hidden_dim == decoder_in_channels[-1]:
                proj_layer = keras.layers.Identity(
                    name=f"decoder_input_proj_identity_{i}",
                    dtype=dtype,
                )
                self.decoder_input_proj_layers.append(proj_layer)
            else:
                self.decoder_input_proj_layers.append(
                    [
                        keras.layers.Conv2D(
                            filters=hidden_dim,
                            kernel_size=1,
                            use_bias=False,
                            kernel_initializer=initializer,
                            bias_initializer="zeros",
                            data_format=data_format,
                            name=f"decoder_input_proj_conv1_{i}",
                            dtype=dtype,
                        ),
                        keras.layers.BatchNormalization(
                            epsilon=1e-5,
                            axis=channel_axis,
                            name=f"decoder_input_proj_bn1_{i}",
                            dtype=dtype,
                        ),
                    ]
                )
        for i in range(num_feature_levels - num_backbone_outs):
            idx = num_backbone_outs + i
            if hidden_dim == decoder_in_channels[-1]:
                proj_layer = keras.layers.Identity(
                    name=f"decoder_input_proj_identity_{idx}",
                    dtype=dtype,
                )
                self.decoder_input_proj_layers.append(proj_layer)
            else:
                self.decoder_input_proj_layers.append(
                    [
                        keras.layers.Conv2D(
                            filters=hidden_dim,
                            kernel_size=3,
                            strides=2,
                            padding="same",
                            use_bias=False,
                            kernel_initializer=initializer,
                            bias_initializer="zeros",
                            data_format=data_format,
                            name=f"decoder_input_proj_conv3_{idx}",
                            dtype=dtype,
                        ),
                        keras.layers.BatchNormalization(
                            epsilon=1e-5,
                            axis=channel_axis,
                            name=f"decoder_input_proj_bn3_{idx}",
                            dtype=dtype,
                        ),
                    ]
                )
        self.dn_split_point = None

        # === Functional Model ===
        pixel_values = keras.Input(
            shape=image_shape, name="pixel_values", dtype="float32"
        )
        feature_maps_output = self.backbone(pixel_values)
        feature_maps = [feature_maps_output[stage] for stage in out_features]
        feature_maps_output_tuple = tuple(feature_maps)
        proj_feats = []
        for level, feature_map in enumerate(feature_maps_output_tuple):
            x = self.encoder_input_proj_layers[level][0](feature_map)
            x = self.encoder_input_proj_layers[level][1](x)
            proj_feats.append(x)
        encoder_outputs = self.encoder(
            inputs_embeds=proj_feats,
            output_hidden_states=True,
            output_attentions=True,
        )
        encoder_last_hidden_state = encoder_outputs[0]
        encoder_hidden_states = (
            encoder_outputs[1] if len(encoder_outputs) > 1 else None
        )
        encoder_attentions = (
            encoder_outputs[2] if len(encoder_outputs) > 2 else None
        )
        last_hidden_state = encoder_outputs[0]
        sources = []
        # NOTE: Handle both no-op (identity mapping) and an actual projection
        # using Conv2D and BatchNorm with `isinstance(proj, list)`.
        for level, source in enumerate(last_hidden_state):
            proj = self.decoder_input_proj_layers[level]
            if isinstance(proj, list):
                x = proj[0](source)
                x = proj[1](x)
                sources.append(x)
            else:
                sources.append(proj(source))
        if num_feature_levels > len(sources):
            len_sources = len(sources)
            proj = self.decoder_input_proj_layers[len_sources]
            if isinstance(proj, list):
                x = proj[0](last_hidden_state[-1])
                x = proj[1](x)
                sources.append(x)
            else:
                sources.append(proj(last_hidden_state[-1]))
            for i in range(len_sources + 1, num_feature_levels):
                proj = self.decoder_input_proj_layers[i]
                if isinstance(proj, list):
                    x = proj[0](sources[-1])
                    x = proj[1](x)
                    sources.append(x)
                else:
                    sources.append(proj(sources[-1]))
        spatial_shapes_tensor = self.spatial_shapes_extractor(sources)
        source_flatten = self.source_flattener(sources)
        if num_denoising > 0 and labels is not None:
            (
                input_query_class,
                denoising_bbox_unact,
                attention_mask,
                denoising_meta_values,
            ) = self.contrastive_denoising_group_generator(
                targets=labels,
                num_queries=num_queries,
            )
            self.dn_split_point = int(denoising_meta_values["dn_num_split"][0])
        else:
            (
                denoising_class,
                denoising_bbox_unact,
                attention_mask,
                denoising_meta_values,
            ) = None, None, None, None

        if num_denoising > 0 and labels is not None:
            denoising_processor = DFineDenoisingPreprocessorLayer(
                name="denoising_processor", dtype=dtype
            )
            denoising_tensors = denoising_processor(
                [
                    pixel_values,
                    input_query_class,
                    denoising_bbox_unact,
                    attention_mask,
                ],
                denoising_meta_values=denoising_meta_values,
            )
            input_query_class_tensor = denoising_tensors["input_query_class"]
            denoising_bbox_unact = denoising_tensors["denoising_bbox_unact"]
            attention_mask = denoising_tensors["attention_mask"]
            denoising_class = self.denoising_class_embed(
                input_query_class_tensor
            )

        anchors, valid_mask = self.anchor_generator(sources)
        memory = keras.ops.where(valid_mask, source_flatten, 0.0)
        output_memory = self.enc_output_layers[0](memory)
        output_memory = self.enc_output_layers[1](output_memory)
        enc_outputs_class = self.enc_score_head(output_memory)
        enc_outputs_coord_logits = self.enc_bbox_head(output_memory)
        enc_outputs_coord_logits_plus_anchors = (
            enc_outputs_coord_logits + anchors
        )
        init_reference_points, target, enc_topk_logits, enc_topk_bboxes = (
            self.initial_query_reference_generator(
                (
                    enc_outputs_class,
                    enc_outputs_coord_logits_plus_anchors,
                    output_memory,
                    sources[-1],
                ),
                denoising_bbox_unact=denoising_bbox_unact,
                denoising_class=denoising_class,
            )
        )
        decoder_outputs = self.decoder(
            inputs_embeds=target,
            encoder_hidden_states=source_flatten,
            reference_points=init_reference_points,
            spatial_shapes=spatial_shapes_tensor,
            attention_mask=attention_mask,
            output_hidden_states=True,
            output_attentions=True,
        )
        last_hidden_state = decoder_outputs[0]
        intermediate_hidden_states = decoder_outputs[1]
        intermediate_logits = decoder_outputs[2]
        intermediate_reference_points = decoder_outputs[3]
        intermediate_predicted_corners = decoder_outputs[4]
        initial_reference_points = decoder_outputs[5]
        decoder_hidden_states = (
            decoder_outputs[6] if len(decoder_outputs) > 6 else None
        )
        decoder_attentions = (
            decoder_outputs[7] if len(decoder_outputs) > 7 else None
        )
        cross_attentions = (
            decoder_outputs[8] if len(decoder_outputs) > 8 else None
        )
        outputs = {
            "last_hidden_state": last_hidden_state,
            "intermediate_hidden_states": intermediate_hidden_states,
            "intermediate_logits": intermediate_logits,
            "intermediate_reference_points": intermediate_reference_points,
            "intermediate_predicted_corners": intermediate_predicted_corners,
            "initial_reference_points": initial_reference_points,
            "decoder_hidden_states": decoder_hidden_states,
            "decoder_attentions": decoder_attentions,
            "cross_attentions": cross_attentions,
            "encoder_last_hidden_state": encoder_last_hidden_state[0],
            "encoder_hidden_states": encoder_hidden_states,
            "encoder_attentions": encoder_attentions,
            "init_reference_points": init_reference_points,
            "enc_topk_logits": enc_topk_logits,
            "enc_topk_bboxes": enc_topk_bboxes,
            "enc_outputs_class": enc_outputs_class,
            "enc_outputs_coord_logits": enc_outputs_coord_logits,
        }

        if num_denoising > 0 and labels is not None:
            outputs["dn_positive_idx"] = denoising_tensors["dn_positive_idx"]
            outputs["dn_num_group"] = denoising_tensors["dn_num_group"]
            outputs["dn_num_split"] = denoising_tensors["dn_num_split"]

        outputs = {k: v for k, v in outputs.items() if v is not None}
        super().__init__(
            inputs=pixel_values,
            outputs=outputs,
            dtype=dtype,
            **kwargs,
        )

        # === Config ===
        self.decoder_in_channels = decoder_in_channels
        self.encoder_hidden_dim = encoder_hidden_dim
        self.num_labels = num_labels
        self.num_denoising = num_denoising
        self.learn_initial_query = learn_initial_query
        self.num_queries = num_queries
        self.anchor_image_size = anchor_image_size
        self.feat_strides = feat_strides
        self.num_feature_levels = num_feature_levels
        self.hidden_dim = hidden_dim
        self.encoder_in_channels = encoder_in_channels
        self.encode_proj_layers = encode_proj_layers
        self.num_attention_heads = num_attention_heads
        self.encoder_ffn_dim = encoder_ffn_dim
        self.num_encoder_layers = num_encoder_layers
        self.hidden_expansion = hidden_expansion
        self.depth_multiplier = depth_multiplier
        self.eval_idx = eval_idx
        self.box_noise_scale = box_noise_scale
        self.labels = labels
        self.label_noise_ratio = label_noise_ratio
        self.num_decoder_layers = num_decoder_layers
        self.decoder_attention_heads = decoder_attention_heads
        self.decoder_ffn_dim = decoder_ffn_dim
        self.decoder_method = decoder_method
        self.decoder_n_points = decoder_n_points
        self.lqe_hidden_dim = lqe_hidden_dim
        self.num_lqe_layers = num_lqe_layers
        self.data_format = data_format
        self.seed = seed
        self.image_shape = image_shape
        self.channel_axis = channel_axis
        self.spatial_shapes = spatial_shapes
        self.stage_names = stage_names
        self.out_features = out_features
        self.initializer = initializer

    def get_config(self):
        config = super().get_config()
        serializable_labels = None
        if self.labels is not None:
            serializable_labels = []
            for target in self.labels:
                serializable_target = {}
                for key, value in target.items():
                    if hasattr(value, "tolist"):
                        serializable_target[key] = value.tolist()
                    else:
                        serializable_target[key] = value
                serializable_labels.append(serializable_target)
        config.update(
            {
                "backbone": keras.layers.serialize(self.backbone),
                "decoder_in_channels": self.decoder_in_channels,
                "encoder_hidden_dim": self.encoder_hidden_dim,
                "num_labels": self.num_labels,
                "num_denoising": self.num_denoising,
                "learn_initial_query": self.learn_initial_query,
                "num_queries": self.num_queries,
                "anchor_image_size": self.anchor_image_size,
                "feat_strides": self.feat_strides,
                "num_feature_levels": self.num_feature_levels,
                "hidden_dim": self.hidden_dim,
                "encoder_in_channels": self.encoder_in_channels,
                "encode_proj_layers": self.encode_proj_layers,
                "num_attention_heads": self.num_attention_heads,
                "encoder_ffn_dim": self.encoder_ffn_dim,
                "num_encoder_layers": self.num_encoder_layers,
                "hidden_expansion": self.hidden_expansion,
                "depth_multiplier": self.depth_multiplier,
                "eval_idx": self.eval_idx,
                "box_noise_scale": self.box_noise_scale,
                "label_noise_ratio": self.label_noise_ratio,
                "labels": serializable_labels,
                "num_decoder_layers": self.num_decoder_layers,
                "decoder_attention_heads": self.decoder_attention_heads,
                "decoder_ffn_dim": self.decoder_ffn_dim,
                "decoder_method": self.decoder_method,
                "decoder_n_points": self.decoder_n_points,
                "lqe_hidden_dim": self.lqe_hidden_dim,
                "num_lqe_layers": self.num_lqe_layers,
                "seed": self.seed,
                "image_shape": self.image_shape,
                "data_format": self.data_format,
                "out_features": self.out_features,
            }
        )
        return config

    @classmethod
    def from_config(cls, config, custom_objects=None):
        config = config.copy()
        if "labels" in config and config["labels"] is not None:
            labels = config["labels"]
            deserialized_labels = []
            for target in labels:
                deserialized_target = {}
                for key, value in target.items():
                    if isinstance(value, list):
                        deserialized_target[key] = np.array(value)
                    else:
                        deserialized_target[key] = value
                deserialized_labels.append(deserialized_target)
            config["labels"] = deserialized_labels
        if "dtype" in config and config["dtype"] is not None:
            dtype_config = config["dtype"]
            if "dtype" not in config["backbone"]["config"]:
                config["backbone"]["config"]["dtype"] = dtype_config
        config["backbone"] = keras.layers.deserialize(
            config["backbone"], custom_objects=custom_objects
        )
        return cls(**config)



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.deit.deit_backbone import DeiTBackbone
from keras_hub.src.models.deit.deit_image_converter import DeiTImageConverter
from keras_hub.src.models.image_classifier_preprocessor import (
    ImageClassifierPreprocessor,
)


@keras_hub_export("keras_hub.models.DeiTImageClassifierPreprocessor")
class DeiTImageClassifierPreprocessor(ImageClassifierPreprocessor):
    backbone_cls = DeiTBackbone
    image_converter_cls = DeiTImageConverter



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.hgnetv2.hgnetv2_encoder import HGNetV2Encoder
from keras_hub.src.models.hgnetv2.hgnetv2_layers import HGNetV2Embeddings
from keras_hub.src.utils.keras_utils import standardize_data_format


@keras_hub_export("keras_hub.models.HGNetV2Backbone")
class HGNetV2Backbone(Backbone):
    """This class represents a Keras Backbone of the HGNetV2 model.

    This class implements an HGNetV2 backbone architecture, a convolutional
    neural network (CNN) optimized for GPU efficiency. HGNetV2 is frequently
    used as a lightweight CNN backbone in object detection pipelines like
    RT-DETR and YOLO variants, delivering strong performance on classification
    and detection tasks, with speed-ups and accuracy gains compared to larger
    CNN backbones.

    Args:
        depths: list of ints, the number of blocks in each stage.
        embedding_size: int, the size of the embedding layer.
        hidden_sizes: list of ints, the sizes of the hidden layers.
        stem_channels: list of ints, the channels for the stem part.
        hidden_act: str, the activation function for hidden layers.
        use_learnable_affine_block: bool, whether to use learnable affine
            transformations.
        stackwise_stage_filters: list of tuples, where each tuple contains
            configuration for a stage: (stage_in_channels, stage_mid_channels,
            stage_out_channels, stage_num_blocks, stage_num_of_layers,
            stage_kernel_size).
            - stage_in_channels: int, input channels for the stage
            - stage_mid_channels: int, middle channels for the stage
            - stage_out_channels: int, output channels for the stage
            - stage_num_blocks: int, number of blocks in the stage
            - stage_num_of_layers: int, number of layers in each block
            - stage_kernel_size: int, kernel size for the stage
        apply_downsample: list of bools, whether to downsample in each stage.
        use_lightweight_conv_block: list of bools, whether to use HGNetV2
            lightweight convolutional blocks in each stage.
        image_shape: tuple, the shape of the input image without the batch size.
            Defaults to `(None, None, 3)`.
        data_format: `None` or str, the data format ('channels_last' or
            'channels_first'). If not specified, defaults to the
            `image_data_format` value in your Keras config.
        out_features: list of str or `None`, the names of the output features to
            return. If `None`, returns all available features from all stages.
            Defaults to `None`.
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`, the data
            type for computations and weights.

    Examples:
    ```python
    import numpy as np
    from keras_hub.src.models.hgnetv2.hgnetv2_backbone import HGNetV2Backbone
    input_data = np.ones(shape=(8, 224, 224, 3))

    # Pretrained backbone.
    model = keras_hub.models.HGNetV2Backbone.from_preset(
        "hgnetv2_b5_ssld_stage2_ft_in1k"
    )
    model(input_data)

    # Randomly initialized backbone with a custom config.
    model = HGNetV2Backbone(
        depths=[1, 2, 4],
        embedding_size=32,
        hidden_sizes=[64, 128, 256],
        stem_channels=[3, 16, 32],
        hidden_act="relu",
        use_learnable_affine_block=False,
        stackwise_stage_filters=[
            (32, 16, 64, 1, 1, 3),     # Stage 0
            (64, 32, 128, 2, 1, 3),    # Stage 1
            (128, 64, 256, 4, 1, 3),   # Stage 2
        ],
        apply_downsample=[False, True, True],
        use_lightweight_conv_block=[False, False, False],
        image_shape=(224, 224, 3),
    )
    model(input_data)
    ```
    """

    def __init__(
        self,
        depths,
        embedding_size,
        hidden_sizes,
        stem_channels,
        hidden_act,
        use_learnable_affine_block,
        stackwise_stage_filters,
        apply_downsample,
        use_lightweight_conv_block,
        image_shape=(None, None, 3),
        data_format=None,
        out_features=None,
        dtype=None,
        **kwargs,
    ):
        name = kwargs.get("name", None)
        data_format = standardize_data_format(data_format)
        channel_axis = -1 if data_format == "channels_last" else 1
        self.image_shape = image_shape
        (
            stage_in_channels,
            stage_mid_channels,
            stage_out_filters,
            stage_num_blocks,
            stage_num_of_layers,
            stage_kernel_size,
        ) = zip(*stackwise_stage_filters)

        # === Layers ===
        self.embedder_layer = HGNetV2Embeddings(
            stem_channels=stem_channels,
            hidden_act=hidden_act,
            use_learnable_affine_block=use_learnable_affine_block,
            data_format=data_format,
            channel_axis=channel_axis,
            name=f"{name}_embedder" if name else "embedder",
            dtype=dtype,
        )
        self.encoder_layer = HGNetV2Encoder(
            stage_in_channels=stage_in_channels,
            stage_mid_channels=stage_mid_channels,
            stage_out_channels=stage_out_filters,
            stage_num_blocks=stage_num_blocks,
            stage_num_of_layers=stage_num_of_layers,
            apply_downsample=apply_downsample,
            use_lightweight_conv_block=use_lightweight_conv_block,
            stage_kernel_size=stage_kernel_size,
            use_learnable_affine_block=use_learnable_affine_block,
            data_format=data_format,
            channel_axis=channel_axis,
            name=f"{name}_encoder" if name else "encoder",
            dtype=dtype,
        )
        self.stage_names = ["stem"] + [
            f"stage{i + 1}" for i in range(len(stackwise_stage_filters))
        ]
        self.out_features = (
            out_features if out_features is not None else self.stage_names
        )

        # === Functional Model ===
        pixel_values = keras.layers.Input(
            shape=image_shape, name="pixel_values_input"
        )
        embedding_output = self.embedder_layer(pixel_values)
        all_encoder_hidden_states_tuple = self.encoder_layer(embedding_output)
        feature_maps_output = {
            stage_name: all_encoder_hidden_states_tuple[idx]
            for idx, stage_name in enumerate(self.stage_names)
            if stage_name in self.out_features
        }
        super().__init__(
            inputs=pixel_values,
            outputs=feature_maps_output,
            dtype=dtype,
            **kwargs,
        )

        # === Config ===
        self.depths = depths
        self.embedding_size = embedding_size
        self.hidden_sizes = hidden_sizes
        self.stem_channels = stem_channels
        self.hidden_act = hidden_act
        self.use_learnable_affine_block = use_learnable_affine_block
        self.stackwise_stage_filters = stackwise_stage_filters
        self.apply_downsample = apply_downsample
        self.use_lightweight_conv_block = use_lightweight_conv_block
        self.data_format = data_format

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "depths": self.depths,
                "embedding_size": self.embedding_size,
                "hidden_sizes": self.hidden_sizes,
                "stem_channels": self.stem_channels,
                "hidden_act": self.hidden_act,
                "use_learnable_affine_block": self.use_learnable_affine_block,
                "stackwise_stage_filters": self.stackwise_stage_filters,
                "apply_downsample": self.apply_downsample,
                "use_lightweight_conv_block": self.use_lightweight_conv_block,
                "image_shape": self.image_shape,
                "out_features": self.out_features,
                "data_format": self.data_format,
            }
        )
        return config



from keras_hub.src.models.gemma3.gemma3_backbone import Gemma3Backbone
from keras_hub.src.models.gemma3.gemma3_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, Gemma3Backbone)



from keras_hub.src.models.depth_anything.depth_anything_backbone import (
    DepthAnythingBackbone,
)
from keras_hub.src.models.depth_anything.depth_anything_presets import (
    backbone_presets,
)
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, DepthAnythingBackbone)
